from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Tuple, Sequence, Text

from pyannote.core import SlidingWindowFeature
from pyannote.metrics.base import BaseMetric

from .. import utils
from ..audio import FilePath, AudioLoader


@dataclass
class HyperParameter:
    """Represents a pipeline hyper-parameter that can be tuned by diart"""

    name: Text
    """Name of the hyper-parameter (e.g. tau_active)"""
    low: float
    """Lowest value that this parameter can take"""
    high: float
    """Highest value that this parameter can take"""

    @staticmethod
    def from_name(name: Text) -> "HyperParameter":
        """Create a HyperParameter object given its name.

        Parameters
        ----------
        name: str
            Name of the hyper-parameter

        Returns
        -------
        HyperParameter
        """
        if name == "tau_active":
            return TauActive
        if name == "rho_update":
            return RhoUpdate
        if name == "delta_new":
            return DeltaNew
        raise ValueError(f"Hyper-parameter '{name}' not recognized")


TauActive = HyperParameter("tau_active", low=0, high=1)
RhoUpdate = HyperParameter("rho_update", low=0, high=1)
DeltaNew = HyperParameter("delta_new", low=0, high=2)


class PipelineConfig(ABC):
    """Configuration containing the required
    parameters to build and run a pipeline"""

    @property
    @abstractmethod
    def duration(self) -> float:
        """The duration of an input audio chunk (in seconds)"""
        pass

    @property
    @abstractmethod
    def step(self) -> float:
        """The step between two consecutive input audio chunks (in seconds)"""
        pass

    @property
    @abstractmethod
    def latency(self) -> float:
        """The algorithmic latency of the pipeline (in seconds).
        At time `t` of the audio stream, the pipeline will
        output predictions for time `t - latency`.
        """
        pass

    @property
    @abstractmethod
    def sample_rate(self) -> int:
        """The sample rate of the input audio stream"""
        pass

    def get_file_padding(self, filepath: FilePath) -> Tuple[float, float]:
        file_duration = AudioLoader(self.sample_rate, mono=True).get_duration(filepath)
        right = utils.get_padding_right(self.latency, self.step)
        left = utils.get_padding_left(file_duration + right, self.duration)
        return left, right


class Pipeline(ABC):
    """Represents a streaming audio pipeline"""

    @staticmethod
    @abstractmethod
    def get_config_class() -> type:
        pass

    @staticmethod
    @abstractmethod
    def suggest_metric() -> BaseMetric:
        pass

    @staticmethod
    @abstractmethod
    def hyper_parameters() -> Sequence[HyperParameter]:
        pass

    @property
    @abstractmethod
    def config(self) -> PipelineConfig:
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def set_timestamp_shift(self, shift: float):
        pass

    @abstractmethod
    def __call__(
        self, waveforms: Sequence[SlidingWindowFeature]
    ) -> Sequence[Tuple[Any, SlidingWindowFeature]]:
        """Runs the next steps of the pipeline
        given a list of consecutive audio chunks.

        Parameters
        ----------
        waveforms: Sequence[SlidingWindowFeature]
            Consecutive chunk waveforms for the pipeline to ingest

        Returns
        -------
        Sequence[Tuple[Any, SlidingWindowFeature]]
            For each input waveform, a tuple containing
            the pipeline output and its respective audio
        """
        pass



from __future__ import annotations

import torch


def overlapped_speech_penalty(
    segmentation: torch.Tensor, gamma: float = 3, beta: float = 10
):
    # segmentation has shape (batch, frames, speakers)
    probs = torch.softmax(beta * segmentation, dim=-1)
    weights = torch.pow(segmentation, gamma) * torch.pow(probs, gamma)
    weights[weights < 1e-8] = 1e-8
    return weights


def normalize_embeddings(
    embeddings: torch.Tensor, norm: float | torch.Tensor = 1
) -> torch.Tensor:
    # embeddings has shape (batch, speakers, feat) or (speakers, feat)
    if embeddings.ndim == 2:
        embeddings = embeddings.unsqueeze(0)
    if isinstance(norm, torch.Tensor):
        batch_size1, num_speakers1, _ = norm.shape
        batch_size2, num_speakers2, _ = embeddings.shape
        assert batch_size1 == batch_size2 and num_speakers1 == num_speakers2
    emb_norm = torch.norm(embeddings, p=2, dim=-1, keepdim=True)
    return norm * embeddings / emb_norm



from __future__ import annotations

from typing import Sequence

import numpy as np
import torch
from pyannote.core import (
    Annotation,
    Timeline,
    SlidingWindowFeature,
    SlidingWindow,
    Segment,
)
from pyannote.metrics.base import BaseMetric
from pyannote.metrics.detection import DetectionErrorRate
from typing_extensions import Literal

from . import base
from .aggregation import DelayedAggregation
from .segmentation import SpeakerSegmentation
from .utils import Binarize
from .. import models as m
from .. import utils


class VoiceActivityDetectionConfig(base.PipelineConfig):
    def __init__(
        self,
        segmentation: m.SegmentationModel | None = None,
        duration: float = 5,
        step: float = 0.5,
        latency: float | Literal["max", "min"] | None = None,
        tau_active: float = 0.6,
        device: torch.device | None = None,
        sample_rate: int = 16000,
        **kwargs,
    ):
        # Default segmentation model is pyannote/segmentation
        self.segmentation = segmentation or m.SegmentationModel.from_pyannote(
            "pyannote/segmentation"
        )

        self._duration = duration
        self._step = step
        self._sample_rate = sample_rate

        # Latency defaults to the step duration
        self._latency = latency
        if self._latency is None or self._latency == "min":
            self._latency = self._step
        elif self._latency == "max":
            self._latency = self._duration

        self.tau_active = tau_active
        self.device = device or torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )

    @property
    def duration(self) -> float:
        return self._duration

    @property
    def step(self) -> float:
        return self._step

    @property
    def latency(self) -> float:
        return self._latency

    @property
    def sample_rate(self) -> int:
        return self._sample_rate


class VoiceActivityDetection(base.Pipeline):
    def __init__(self, config: VoiceActivityDetectionConfig | None = None):
        self._config = VoiceActivityDetectionConfig() if config is None else config

        msg = f"Latency should be in the range [{self._config.step}, {self._config.duration}]"
        assert self._config.step <= self._config.latency <= self._config.duration, msg

        self.segmentation = SpeakerSegmentation(
            self._config.segmentation, self._config.device
        )
        self.pred_aggregation = DelayedAggregation(
            self._config.step,
            self._config.latency,
            strategy="hamming",
            cropping_mode="loose",
        )
        self.audio_aggregation = DelayedAggregation(
            self._config.step,
            self._config.latency,
            strategy="first",
            cropping_mode="center",
        )
        self.binarize = Binarize(self._config.tau_active)

        # Internal state, handle with care
        self.timestamp_shift = 0
        self.chunk_buffer, self.pred_buffer = [], []

    @staticmethod
    def get_config_class() -> type:
        return VoiceActivityDetectionConfig

    @staticmethod
    def suggest_metric() -> BaseMetric:
        return DetectionErrorRate(collar=0, skip_overlap=False)

    @staticmethod
    def hyper_parameters() -> Sequence[base.HyperParameter]:
        return [base.TauActive]

    @property
    def config(self) -> base.PipelineConfig:
        return self._config

    def reset(self):
        self.set_timestamp_shift(0)
        self.chunk_buffer, self.pred_buffer = [], []

    def set_timestamp_shift(self, shift: float):
        self.timestamp_shift = shift

    def __call__(
        self,
        waveforms: Sequence[SlidingWindowFeature],
    ) -> Sequence[tuple[Annotation, SlidingWindowFeature]]:
        batch_size = len(waveforms)
        msg = "Pipeline expected at least 1 input"
        assert batch_size >= 1, msg

        # Create batch from chunk sequence, shape (batch, samples, channels)
        batch = torch.stack([torch.from_numpy(w.data) for w in waveforms])

        expected_num_samples = int(
            np.rint(self.config.duration * self.config.sample_rate)
        )
        msg = f"Expected {expected_num_samples} samples per chunk, but got {batch.shape[1]}"
        assert batch.shape[1] == expected_num_samples, msg

        # Extract segmentation
        segmentations = self.segmentation(batch)  # shape (batch, frames, speakers)
        voice_detection = torch.max(segmentations, dim=-1, keepdim=True)[
            0
        ]  # shape (batch, frames, 1)

        seg_resolution = waveforms[0].extent.duration / segmentations.shape[1]

        outputs = []
        for wav, vad in zip(waveforms, voice_detection):
            # Add timestamps to segmentation
            sw = SlidingWindow(
                start=wav.extent.start,
                duration=seg_resolution,
                step=seg_resolution,
            )
            vad = SlidingWindowFeature(vad.cpu().numpy(), sw)

            # Update sliding buffer
            self.chunk_buffer.append(wav)
            self.pred_buffer.append(vad)

            # Aggregate buffer outputs for this time step
            agg_waveform = self.audio_aggregation(self.chunk_buffer)
            agg_prediction = self.pred_aggregation(self.pred_buffer)
            agg_prediction = self.binarize(agg_prediction).get_timeline(copy=False)

            # Shift prediction timestamps if required
            if self.timestamp_shift != 0:
                shifted_agg_prediction = Timeline(uri=agg_prediction.uri)
                for segment in agg_prediction:
                    new_segment = Segment(
                        segment.start + self.timestamp_shift,
                        segment.end + self.timestamp_shift,
                    )
                    shifted_agg_prediction.add(new_segment)
                agg_prediction = shifted_agg_prediction

            # Convert timeline into annotation with single speaker "speech"
            agg_prediction = agg_prediction.to_annotation(utils.repeat_label("speech"))
            outputs.append((agg_prediction, agg_waveform))

            # Make place for new chunks in buffer if required
            if len(self.chunk_buffer) == self.pred_aggregation.num_overlapping_windows:
                self.chunk_buffer = self.chunk_buffer[1:]
                self.pred_buffer = self.pred_buffer[1:]

        return outputs



from __future__ import annotations

from abc import ABC
from pathlib import Path
from typing import Optional, Text, Union, Callable, List

import numpy as np
import torch
import torch.nn as nn
from requests import HTTPError

try:
    from pyannote.audio import Model
    from pyannote.audio.pipelines.speaker_verification import PretrainedSpeakerEmbedding
    from pyannote.audio.utils.powerset import Powerset

    IS_PYANNOTE_AVAILABLE = True
except ImportError:
    IS_PYANNOTE_AVAILABLE = False

try:
    import onnxruntime as ort

    IS_ONNX_AVAILABLE = True
except ImportError:
    IS_ONNX_AVAILABLE = False


class PowersetAdapter(nn.Module):
    def __init__(self, segmentation_model: nn.Module):
        super().__init__()
        self.model = segmentation_model
        specs = self.model.specifications
        max_speakers_per_frame = specs.powerset_max_classes
        max_speakers_per_chunk = len(specs.classes)
        self.powerset = Powerset(max_speakers_per_chunk, max_speakers_per_frame)

    def forward(self, waveform: torch.Tensor) -> torch.Tensor:
        return self.powerset.to_multilabel(self.model(waveform))


class PyannoteLoader:
    def __init__(self, model_info, hf_token: Union[Text, bool, None] = True):
        super().__init__()
        self.model_info = model_info
        self.hf_token = hf_token

    def __call__(self) -> Callable:
        try:
            model = Model.from_pretrained(self.model_info, use_auth_token=self.hf_token)
            specs = getattr(model, "specifications", None)
            if specs is not None and specs.powerset:
                model = PowersetAdapter(model)
            return model
        except HTTPError:
            pass
        except ModuleNotFoundError:
            pass
        return PretrainedSpeakerEmbedding(self.model_info, use_auth_token=self.hf_token)


class ONNXLoader:
    def __init__(self, path: str | Path, input_names: List[str], output_name: str):
        super().__init__()
        self.path = Path(path)
        self.input_names = input_names
        self.output_name = output_name

    def __call__(self) -> ONNXModel:
        return ONNXModel(self.path, self.input_names, self.output_name)


class ONNXModel:
    def __init__(self, path: Path, input_names: List[str], output_name: str):
        super().__init__()
        self.path = path
        self.input_names = input_names
        self.output_name = output_name
        self.device = torch.device("cpu")
        self.session = None
        self.recreate_session()

    @property
    def execution_provider(self) -> str:
        device = "CUDA" if self.device.type == "cuda" else "CPU"
        return f"{device}ExecutionProvider"

    def recreate_session(self):
        options = ort.SessionOptions()
        options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
        self.session = ort.InferenceSession(
            self.path,
            sess_options=options,
            providers=[self.execution_provider],
        )

    def to(self, device: torch.device) -> ONNXModel:
        if device.type != self.device.type:
            self.device = device
            self.recreate_session()
        return self

    def __call__(self, *args) -> torch.Tensor:
        inputs = {
            name: arg.cpu().numpy().astype(np.float32)
            for name, arg in zip(self.input_names, args)
        }
        output = self.session.run([self.output_name], inputs)[0]
        return torch.from_numpy(output).float().to(args[0].device)


class LazyModel(ABC):
    def __init__(self, loader: Callable[[], Callable]):
        super().__init__()
        self.get_model = loader
        self.model: Optional[Callable] = None

    def is_in_memory(self) -> bool:
        """Return whether the model has been loaded into memory"""
        return self.model is not None

    def load(self):
        if not self.is_in_memory():
            self.model = self.get_model()

    def to(self, device: torch.device) -> LazyModel:
        self.load()
        self.model = self.model.to(device)
        return self

    def __call__(self, *args, **kwargs):
        self.load()
        return self.model(*args, **kwargs)

    def eval(self) -> LazyModel:
        self.load()
        if isinstance(self.model, nn.Module):
            self.model.eval()
        return self


class SegmentationModel(LazyModel):
    """
    Minimal interface for a segmentation model.
    """

    @staticmethod
    def from_pyannote(
        model, use_hf_token: Union[Text, bool, None] = True
    ) -> "SegmentationModel":
        """
        Returns a `SegmentationModel` wrapping a pyannote model.

        Parameters
        ----------
        model: pyannote.PipelineModel
            The pyannote.audio model to fetch.
        use_hf_token: str | bool, optional
            The Huggingface access token to use when downloading the model.
            If True, use huggingface-cli login token.
            Defaults to None.

        Returns
        -------
        wrapper: SegmentationModel
        """
        assert IS_PYANNOTE_AVAILABLE, "No pyannote.audio installation found"
        return SegmentationModel(PyannoteLoader(model, use_hf_token))

    @staticmethod
    def from_onnx(
        model_path: Union[str, Path],
        input_name: str = "waveform",
        output_name: str = "segmentation",
    ) -> "SegmentationModel":
        assert IS_ONNX_AVAILABLE, "No ONNX installation found"
        return SegmentationModel(ONNXLoader(model_path, [input_name], output_name))

    @staticmethod
    def from_pretrained(
        model, use_hf_token: Union[Text, bool, None] = True
    ) -> "SegmentationModel":
        if isinstance(model, str) or isinstance(model, Path):
            if Path(model).name.endswith(".onnx"):
                return SegmentationModel.from_onnx(model)
        return SegmentationModel.from_pyannote(model, use_hf_token)

    def __call__(self, waveform: torch.Tensor) -> torch.Tensor:
        """
        Call the forward pass of the segmentation model.
        Parameters
        ----------
        waveform: torch.Tensor, shape (batch, channels, samples)
        Returns
        -------
        speaker_segmentation: torch.Tensor, shape (batch, frames, speakers)
        """
        return super().__call__(waveform)


class EmbeddingModel(LazyModel):
    """Minimal interface for an embedding model."""

    @staticmethod
    def from_pyannote(
        model, use_hf_token: Union[Text, bool, None] = True
    ) -> "EmbeddingModel":
        """
        Returns an `EmbeddingModel` wrapping a pyannote model.

        Parameters
        ----------
        model: pyannote.PipelineModel
            The pyannote.audio model to fetch.
        use_hf_token: str | bool, optional
            The Huggingface access token to use when downloading the model.
            If True, use huggingface-cli login token.
            Defaults to None.

        Returns
        -------
        wrapper: EmbeddingModel
        """
        assert IS_PYANNOTE_AVAILABLE, "No pyannote.audio installation found"
        loader = PyannoteLoader(model, use_hf_token)
        return EmbeddingModel(loader)

    @staticmethod
    def from_onnx(
        model_path: Union[str, Path],
        input_names: List[str] | None = None,
        output_name: str = "embedding",
    ) -> "EmbeddingModel":
        assert IS_ONNX_AVAILABLE, "No ONNX installation found"
        input_names = input_names or ["waveform", "weights"]
        loader = ONNXLoader(model_path, input_names, output_name)
        return EmbeddingModel(loader)

    @staticmethod
    def from_pretrained(
        model, use_hf_token: Union[Text, bool, None] = True
    ) -> "EmbeddingModel":
        if isinstance(model, str) or isinstance(model, Path):
            if Path(model).name.endswith(".onnx"):
                return EmbeddingModel.from_onnx(model)
        return EmbeddingModel.from_pyannote(model, use_hf_token)

    def __call__(
        self, waveform: torch.Tensor, weights: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Call the forward pass of an embedding model with optional weights.
        Parameters
        ----------
        waveform: torch.Tensor, shape (batch, channels, samples)
        weights: Optional[torch.Tensor], shape (batch, frames)
            Temporal weights for each sample in the batch. Defaults to no weights.
        Returns
        -------
        speaker_embeddings: torch.Tensor, shape (batch, embedding_dim)
        """
        embeddings = super().__call__(waveform, weights)
        if isinstance(embeddings, np.ndarray):
            embeddings = torch.from_numpy(embeddings)
        return embeddings
